package eu.jonahbauer.raytracing.render;

import eu.jonahbauer.raytracing.math.Range;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.scene.Scene;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public record Camera(
        int width, int height,
        double viewportWidth, double viewportHeight,
        @NotNull Vec3 origin, @NotNull Vec3 direction
) {
    public Camera {
        if (width <= 0) throw new IllegalArgumentException("width must be positive");
        if (height <= 0) throw new IllegalArgumentException("height must be positive");
        if (viewportWidth <= 0 || !Double.isFinite(viewportWidth)) throw new IllegalArgumentException("viewportWidth must be positive");
        if (viewportHeight <= 0 || !Double.isFinite(viewportHeight)) throw new IllegalArgumentException("viewportHeight must be positive");
        Objects.requireNonNull(origin, "origin");
        Objects.requireNonNull(direction, "direction");
    }

    /**
     * Creates a new camera with the given dimensions at the origin facing towards positive z with a focal length of 1.
     * @param height the image height
     * @param viewportHeight the viewport height
     * @param aspectRatio the aspect ratio
     */
    public Camera(int height, double viewportHeight, double aspectRatio) {
        this((int) (height * aspectRatio), height, viewportHeight * aspectRatio, viewportHeight);
    }

    /**
     * Creates a new camera with the given dimensions at the origin facing towards positive z with a focal length of 1.
     * @param width the image width
     * @param height the image height
     * @param viewportWidth the viewport width
     * @param viewportHeight the viewport height
     */
    public Camera(int width, int height, double viewportWidth, double viewportHeight) {
        this(width, height, viewportWidth, viewportHeight, Vec3.ZERO, Vec3.UNIT_Z);
    }

    public @NotNull Image render(@NotNull Scene scene) {
        var image = new Image(width(), height());

        pixels().forEach(pixel -> {
            var x = pixel.x();
            var y = pixel.y();
            var ray = pixel.ray();

            var result = scene.hit(ray, Range.NON_NEGATIVE);
            if (result.isPresent()) {
                var normal = result.get().normal();
                image.set(x, y, getNormalColor(normal));
            } else {
                image.set(x, y, scene.getSkyboxColor(ray));
            }
        });

        return image;
    }

    private @NotNull Color getNormalColor(@NotNull Vec3 normal) {
        return new Color(
                0.5 * (normal.x() + 1),
                0.5 * (normal.y() + 1),
                0.5 * (normal.z() + 1)
        );
    }

    public @NotNull Stream<Pixel> pixels() {
        // project direction onto xz-plane
        var d = direction.unit();
        var dXZ = direction.withY(0).unit();

        var viewportU = new Vec3(dXZ.z(), 0, - dXZ.x()); // perpendicular to dXZ in xz-plane
        var viewportV = d.cross(viewportU); // perpendicular to viewportU and direction

        viewportU = viewportU.times(viewportWidth); // vector along the width of the viewport
        viewportV = viewportV.times(- viewportHeight); // vector along the height of the viewport

        var pixelU = viewportU.div(width);
        var pixelV = viewportV.div(height);

        var pixel00 = origin.plus(direction)
                .minus(viewportU.div(2)).minus(viewportV.div(2))
                .plus(pixelU.div(2)).plus(pixelV.div(2));

        return IntStream.range(0, width * height).mapToObj(i -> {
            var y = i / width;
            var x = i % width;
            var pixel = pixel00.plus(pixelU.times(x)).plus(pixelV.times(y));
            return new Pixel(x, y, new Ray(origin, pixel.minus(origin)));
        });
    }

    public record Pixel(int x, int y, @NotNull Ray ray) {}
}
